'''Test our Connection'''

from base import BaseTest
from cStringIO import StringIO

from s3po.exceptions import DownloadException


class ConnectionTest(BaseTest):
    '''Test our connection's functionality'''

    def test_round_trip(self):
        '''Can upload a string.'''
        self.conn.upload('bucket', 'key', 'content')
        self.assertEqual(self.conn.download('bucket', 'key'), 'content')

    def test_upload_file(self):
        '''Upload from a file object.'''
        self.conn.upload('bucket', 'key', StringIO('content'))
        self.assertEqual(self.conn.download('bucket', 'key'), 'content')

    def test_upload_filename(self):
        '''Upload by filename.'''
        path = self.tmpfile('upload')
        with open(path, 'w+') as fout:
            fout.write('content')

        self.conn.upload_file('bucket', 'key', path)
        self.assertEqual(self.conn.download('bucket', 'key'), 'content')

    def test_download_file(self):
        '''Can download to a file.'''
        fobj = StringIO()
        self.conn.upload('bucket', 'key', 'content')
        self.conn.download('bucket', 'key', fobj)
        self.assertEqual(fobj.getvalue(), 'content')

    def test_download_filename(self):
        '''Download by filename.'''
        path = self.tmpfile('download')
        self.conn.upload('bucket', 'key', 'content')
        self.conn.download_file('bucket','key', path)
        with open(path) as fin:
            self.assertEqual(fin.read(), 'content')

    def test_download_missing(self):
        '''Raises an exception downloading a missing key.'''
        self.assertRaises(DownloadException, self.conn.download, 'bucket', 'key')

    def test_mock(self):
        '''Mocking swaps out the backend.'''
        original = self.conn.backend
        try:
            with self.conn.mock():
                self.assertNotEqual(original, self.conn.backend)
                raise ValueError('Explode')
        except ValueError:
            pass

        self.assertEqual(original, self.conn.backend)

    def test_list(self):
        self.conn.upload('bucket', 'key', StringIO('content'))
        self.assertEqual(list(self.conn.list('bucket')), 
                         ['key'])

    def test_list_prefix(self):
        self.conn.upload('bucket', 'key', StringIO('content'))
        self.conn.upload('bucket', 'something_else', StringIO('content'))
        self.assertEqual(list(self.conn.list('bucket', prefix='k')), 
                         ['key'])

    def test_list_delimiter(self):
        self.conn.upload('bucket', 'a.1', StringIO('content'))
        self.conn.upload('bucket', 'a.2', StringIO('content'))
        self.assertEqual(list(self.conn.list('bucket', delimiter='.')), 
                         ['a'])




'''An in-memory backend'''

from ..exceptions import DownloadException


class Memory(object):
    '''An in-memory backend'''

    def __init__(self):
        self.buckets = {}

    def download(self, bucket, key, fobj, retries, headers=None):
        '''Download the contents of bucket/key to fobj'''
        obj = self.buckets.get(bucket, {}).get(key)
        if not obj:
            raise DownloadException('%s / %s not fount' % (bucket, key))
        else:
            fobj.write(obj)

    def upload(self, bucket, key, fobj, retries, headers=None):
        '''Upload the contents of fobj to bucket/key with headers'''
        if bucket not in self.buckets:
            self.buckets[bucket] = {}

        self.buckets[bucket][key] = fobj.read()

    def list(self, bucket, prefix=None, delimiter=None, retries=None, headers=None):
        '''List the contents of a bucket.'''
        if prefix is None:
            prefix = ''
        keys = (key for key in self.buckets[bucket].keys() if key.startswith(prefix))
        if delimiter:
            return (prefix for prefix in set(key.split(delimiter, 1)[0] for key in keys))
        else:
            return keys



'''Our internal exceptions'''


class S3POException(Exception):
    '''All of S3PO's exceptions'''
    pass


class DownloadException(S3POException):
    '''An error while downloading'''
    pass


class UploadException(S3POException):
    '''An error while uploading'''
    pass



'''Make sure our utitilies work as advertised'''

import unittest

from s3po.util import retry


class UtilTest(unittest.TestCase):
    '''Test our utilities'''
    def test_retry(self):
        '''We should be able to retry functions'''
        @retry(10, sleep=lambda _: 0)
        def func(obj):
            '''Counting function'''
            obj['count'] += 1
            raise ValueError('foo')

        obj = {'count': 0}
        self.assertRaises(ValueError, func, obj)
        self.assertEqual(obj['count'], 10)

    def test_exceptions(self):
        '''We should be able to limit the exceptions we retry on'''
        @retry(10, exceptions=(TypeError))
        def func(obj):
            '''Raise a ValueError'''
            obj['count'] += 1
            raise ValueError('foo')

        obj = {'count': 0}
        self.assertRaises(ValueError, func, obj)
        self.assertEqual(obj['count'], 1)



'''Test all of our batching utilities'''

import unittest

from s3po.connection import Connection
from s3po.exceptions import DownloadException


class BatchTest(unittest.TestCase):
    '''We should be able to batch some requests out'''
    def setUp(self):
        self.conn = Connection.memory()

    def test_upload(self):
        '''We should be able to upload'''
        keys = ['foo', 'bar', 'baz', 'whiz']
        with self.conn.batch() as batch:
            for key in keys:
                batch.upload('bucket', key, 'foo')

        for key in keys:
            self.assertEqual(self.conn.download('bucket', key), 'foo')

    def test_callback(self):
        '''We can optionally provide a callback'''
        keys = ['foo', 'bar', 'baz', 'whiz']
        with self.conn.batch() as batch:
            for key in keys:
                batch.upload('bucket', key, key)

        results = []
        with self.conn.batch() as batch:
            for key in keys:
                batch.download(
                    'bucket', key, callback=lambda r: results.append(r))
        self.assertEqual(set(results), set(keys))

    def test_results(self):
        '''We can get the results of our batch operations'''
        keys = ['foo', 'bar', 'baz', 'whiz']
        with self.conn.batch() as batch:
            for key in keys:
                batch.upload('bucket', key, key)

        with self.conn.batch() as batch:
            for key in keys:
                batch.download('bucket', key)

        self.assertEqual(set(batch.results()), set(keys))

    def test_success(self):
        '''We can get confirmation of a successful batch.'''
        keys = ['foo', 'bar', 'baz', 'whiz']
        with self.conn.batch() as batch:
            for key in keys:
                batch.upload('bucket', key, key)

        self.assertTrue(batch.success())

    def test_failed(self):
        '''We can get confirmation of a failed batch.'''
        keys = ['foo', 'bar', 'baz', 'whiz']
        with self.conn.batch() as batch:
            for key in keys:
                batch.download('bucket', key)

        self.assertFalse(batch.success())



'''Deal with S3.'''


from boto.s3.connection import S3Connection
from cStringIO import StringIO

from ..util import CountFile, retry, logger
from ..exceptions import UploadException, DownloadException


class S3(object):
    '''Our connection to S3'''
    # How big must a file get before it's multiparted. Also how big the chunks
    # are that we'll read
    chunk_size = 50 * 1024 * 1204
    min_chunk_size = 5 * 1024 * 1024

    def __init__(self, *args, **kwargs):
        self.conn = S3Connection(*args, **kwargs)

    def get_bucket(self, bucket):
        return self.conn.get_bucket(bucket, validate=False)

    def download(self, bucket, key, destination, retries, headers=None):
        '''Download the contents of bucket/key to destination'''
        bucket = self.get_bucket(bucket)
        # Make a file that we'll write into
        destination = CountFile(destination)
        obj = bucket.get_key(key)
        if not obj:
            raise DownloadException('Key %s does not exist in %s' % (
                key, bucket.name))

        # Get its original location so we can go back to it if need be
        offset = destination.tell()

        @retry(retries)
        def func():
            '''The bit that we want to retry'''
            destination.seek(offset)
            obj.get_contents_to_file(destination, headers=headers)
            # Ensure it was downloaded completely
            logger.info(
                'Downloaded %s bytes out of %s' % (destination.count, obj.size))
            if obj.size != destination.count:
                raise DownloadException('Downloaded only %i of %i bytes' % (
                    destination.count, obj.size or 0))
        # With our wrapped function defined, we'll go ahead an invoke it.
        func()

    def upload(self, bucket, key, source, retries, headers=None):
        '''Upload the contents of source to bucket/key with headers'''
        # Make our headers object
        headers = headers or {}
        bucket = self.get_bucket(bucket)
        # We'll read in some data, and if the file appears small enough, we'll
        # upload it in a single go. In order for it to be a valid multipart
        # upload, it needs at least two parts, so we will make sure there are
        # at least enough for two parts before we commit to multipart
        data = source.read(2 * self.chunk_size)
        if len(data) < (2 * self.chunk_size):
            key = bucket.new_key(key)

            @retry(retries)
            def func():
                '''The bit that we want to retry'''
                key.set_contents_from_string(data, headers=headers)
                if key.size != len(data):
                    raise UploadException('Uploaded only %i for %i bytes' % (
                        key.size, len(data)))
                return True
            return func()
        else:
            logger.info('Performing multipart upload')
            # Otherwise, it's a large-enough file that we should multipart
            # upload it. There's a restriction that all parts of a multipart
            # upload must be at least 5MB. Therefore, we should keep uploading
            # chunks as long as the remaining data is 5MB greater than our chunk
            # size. That way we avoid the case where we have a remainder less
            # than this limit
            multi = bucket.initiate_multipart_upload(key, headers=headers)
            count = 1
            while len(data) >= (self.chunk_size + self.min_chunk_size):
                logger.info('Uploading chunk %s', count)
                part = data[0:self.chunk_size]
                retry(
                    retries)(multi.upload_part_from_file)(StringIO(part), count)
                data = (
                    data[self.chunk_size:] +
                    source.read(self.chunk_size))
                count += 1
            # And finally, the last part
            multi.upload_part_from_file(StringIO(data), count)
            multi.complete_upload()
            return True

    def _list_retry(self, retries, bucket, *args, **kwargs):
        @retry(retries)
        def func():
            return bucket.list(*args, **kwargs)
        return func()

    def list(self, bucket, prefix=None, delimiter=None, retries=3, headers=None):
        '''List the bucket, possibly limiting the search with a prefix.'''
        bucket = self.conn.get_bucket(bucket)
        # consume iterator to make a list to keep parity with Swift backend
        return (key.name for key in self._list_retry(retries, bucket,
                                          prefix, delimiter, headers=headers))
